import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.net.*;
import java.io.*;
import java.util.Vector;
import java.util.ListIterator;
import java.util.Hashtable;
import javax.swing.border.*;
import javax.swing.text.*;
import javax.swing.text.html.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import fi.helsinki.cs.gist.*;

class PlaceholderDemo {
    protected JComboBox comboBox;
    protected JEditorPane editorPane;
    protected PlaceholderPane scrollPane;
    protected Hashtable hash;
    protected String currentPage; 

    public void setPage(String s) {
	URL url = null;
	try {
	    url = new URL(s);
	} catch (Exception e) {
	    System.err.println("Couldn't create URL: " + s);
	    return;
	}
	
	try {
	    editorPane.setPage(url);
	    System.out.println("URL set successfully to " + url);
	} catch (IOException e) {
	    System.err.println("Attempted to read a bad URL: " + url);
	}	
    }


    public PlaceholderDemo(String[] args) {

	// The default LookAndFeel is JavaLookAndFeel, set it to Windows
	try {
	    UIManager.setLookAndFeel(
		"com.sun.java.swing.plaf.windows.WindowsLookAndFeel");
	    System.out.println("WindowsLookAndFeel set");
	} catch (Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	} 

	final JFrame f = new JFrame("Bookmark Scroll Bar Demo Browser");

	editorPane = new JEditorPane();
	editorPane.setEditable(false);

	scrollPane = new PlaceholderPane(editorPane);
	scrollPane.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_ALWAYS);
	scrollPane.setPreferredSize(new Dimension(710, 616));

	hash = new Hashtable();

	String[] pages = { "http://java.sun.com/docs/books/tutorial/uiswing/components/button.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/colorchooser.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/combobox.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/dialog.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/editorpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/filechooser.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/frame.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/generaltext.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/intermediate.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/internalframe.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/jcomponent.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/label.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/layeredpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/list.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/menu.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/panel.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/problems.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/progress.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/rootpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/scrollpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/simpletext.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/slider.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/splitpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/tabbedpane.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/table.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/text.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/textapi.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/textfield.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/toolbar.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/tooltip.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/toplevel.html",
			   "http://java.sun.com/docs/books/tutorial/uiswing/components/tree.html" };
	
	for (int i = 0; i < pages.length; i++) 
	    hash.put(pages[i], new Vector());

	JPanel textPanel = new JPanel(new BorderLayout());
	comboBox = new JComboBox(pages);
	Container pane = f.getContentPane();
	pane.setLayout(new BorderLayout());
	pane.add(scrollPane, BorderLayout.CENTER);
	pane.add(textPanel, BorderLayout.NORTH);
	Border border = comboBox.getBorder();
	Border margin = new EmptyBorder(0, 2, 0, 0);
	comboBox.setBorder(new CompoundBorder(border, margin));
	textPanel.setBorder(new EmptyBorder(5, 5, 5, 20));
	textPanel.add(comboBox, BorderLayout.CENTER);

	comboBox.addActionListener(new ActionListener() {
		public void actionPerformed(ActionEvent e) {
		    if (currentPage != null)
			hash.put(currentPage, scrollPane.getVerticalScrollBar().getTags());
		    currentPage = (String) comboBox.getSelectedItem();
		    setPage(currentPage);
		    scrollPane.getVerticalScrollBar().setTags((Vector) hash.get(currentPage));
		}
	});

	comboBox.setSelectedItem("http://java.sun.com/docs/books/tutorial/uiswing/components/tree.html");
	currentPage = (String) comboBox.getSelectedItem(); 
	setPage(currentPage);
	// editorPane.addPropertyChangeListener("page", new CodeShader());
	editorPane.addPropertyChangeListener(new CodeShader());

	f.addWindowListener (new WindowAdapter() {
	    public void windowClosing(WindowEvent e) {
		System.exit(0);
	    }
	});

	f.pack();
	f.show();
    }


    public static void main(String[] args) {
	// Working defaults for command line arguments
	if (args.length < 1) {
	    args[0] = "http://www.cs.helsinki.fi/";
	    if (args.length < 2)
		args[1] = "";
	}
	
	new PlaceholderDemo(args);
    }

    class CodeShader implements PropertyChangeListener {
	Vector shadedAreas;

	public CodeShader() {
	    shadedAreas = new Vector();
	}

	public void propertyChange(PropertyChangeEvent evt) {
	    shadedAreas.removeAllElements();
	    findCodeAreas();
	    scrollPane.getVerticalScrollBar().setShadedAreas(shadedAreas);
	}

	public void findCodeAreas() {
	    Document d = editorPane.getDocument();
	    if (d instanceof HTMLDocument) {
		HTMLDocument hd = (HTMLDocument) d;
		Element re[] = hd.getRootElements();
		int j;
		for (j = 0; j < re.length; j++)
		    if (re[j].getName() == "html")
			break;
		if (re[j].getName() != "html") {
		    System.out.println("Failed to find HTML tag!");
		    return;
		}

		Element html = re[j];
		for (j = 0; j < html.getElementCount(); j++)
		    if (html.getElement(j).getName() == "body")
			break;
		if (html.getElement(j).getName() != "body") {
		    System.out.println("Failed to find BODY tag!");
		    return;
		}

		Element body = html.getElement(j);
		for (j = 0; j < body.getElementCount(); j++) {
		    if (body.getElement(j).getName() == "blockquote") {
			Element block1 = body.getElement(j);
			for (int k = 0; k < block1.getElementCount(); k++) {
			    Element block2 = block1.getElement(k);
			    if (block2.getName() == "blockquote") {
				for (int l = 0; l < block2.getElementCount(); l++) {
				    Element pre = block2.getElement(l);
				    if (pre.getName() == "pre") 
					setShadedArea(pre.getStartOffset(), pre.getEndOffset());
				}
			    }
			}
		    }
		}
	    } else {
		System.out.println("The document isn't HTML");
	    }
	}
	
	public void setShadedArea(int start, int end) {
	    Rectangle s = null;
	    Rectangle e = null;
		
	    try {
		s = editorPane.modelToView(start);
		e = editorPane.modelToView(end);
	    } catch (Exception ex) {
		System.out.println("Badlocationexecption");
	    }
			
	    if (s != null && e != null) 
		shadedAreas.addElement(new Rectangle(0, s.y, 0, e.y - s.y));
	}
    }

}
