/*
 * Tag (c) GIST 1999
 */
  
package fi.helsinki.cs.gist.scroll;

import fi.helsinki.cs.gist.*;
import java.util.*;
import javax.swing.text.*;
import javax.swing.*;

import java.awt.*;
import java.awt.event.*;
import java.awt.dnd.*;
import java.awt.datatransfer.*;
import java.io.*;

import java.beans.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;
import javax.swing.border.*;
 
/**
 * Tag is used to hold tag location and the JTextField
 * component which make up the 'tooltip'. 
 *
 * @author asaura
 * @version 11.9.1999
 *
 */

class Tag extends Window {

    static Color basicBg = new Color(255,255,225);
    
    /** Tag highlight background color */
    static Color hilightBg = new Color(255,255,205);

    /** Tag borders */
    static EmptyBorder margins = new EmptyBorder(0, 5, 0, 5);
    static CompoundBorder basicBorder = new CompoundBorder(new LineBorder(Color.black), margins);
    static CompoundBorder hilightBorder = new CompoundBorder(new LineBorder(Color.black, 2), margins);

    /**
     * Inside this threshold a drag initiated from the tag is only a
     * selection. Outside it the drag is considered as a deletion
     * attempt.
     */
    static int DELTHRESHOLD = 70;
  
    /** ScrollBar (Placeholder) model coordinates. Not to be drawn. */
    int value;
    /** Screen coordinates relative to the TagArea. */
    int drawnY;

    /** Prevents this tag from disappearing, when it's under the cursor **/
    boolean stayVisible = false;

    /** Prevents this tag from disappearing, when it's was just created. **/ 
    boolean newBorn = true;
    
    TagArea tagArea;
    /**
     * Layout needed in focus transfer.
     * @see p
     */
    CardLayout cards;

    /** The editable container of the tag text. */
    JTextField tf;
    /**
     * An AWT textfield for gaining the focus. This is a horrible
     * hack: for some reason transfering the focus from one window
     * (the TagArea) to the Tag does not work, if 1) the tag isn't an
     * AWT window and 2) there does not exist an AWT component in the
     * window, which gets the focus. This is a known bug of Swing.
     * 
     * The fix: an AWT TextField is shown first, the focus is
     * transfered to the tag, the Swing textfield is put on top, and
     * the focus is transfered again.
     */
    TextField p;

    /**
     * A "semaphor" for the edit state. This prevents from sending
     * false start or stop editing events.
     */
    boolean editing;

    /** Drag state information. */
    boolean dragInProgress = false;
    /**
     * Drag state information. If true, the drag has gone out of the
     * magic threshold and its now considered as an attempt to delete 
     * the tag.
     */
    boolean delDrag = false;
    /** The windows which follows the cursor in drag. */
    Window dragWindow = null;

    Window topWindow;


    Tag(String newName, int newValue, TagArea newTagArea, Window topW, boolean interactive) {
	super(topW);
	this.topWindow = topW;
	this.tagArea = newTagArea;
	this.setValue(newValue);

	p = new TextField(1);
	this.cards = new CardLayout();
 	this.setLayout(this.cards);
	if(newName.length() < 2)
	    this.tf = new JTextField(2);
	else
	    this.tf = new JTextField(newName);

	tf.setColumns(0);
	this.tf.setBackground(basicBg);
	this.tf.setBorder(basicBorder);
	
	this.add(this.p, "p");
	this.add(this.tf, "tf");

	// The AWT textfield must be shown first, if the tag is
	// created by the user in order to gain focus.
	if (interactive) {
	    this.cards.show(this,"p");
	} else {
	    this.cards.show(this, "tf");
	}

	this.pack();
	this.show();

	p.addFocusListener(new FocusAdapter() {
		public void focusGained(FocusEvent e) {
		    // The focus isn't completely gained, if the textfield 
		    // isn't shown for a split of a second.
		    System.out.println("Got focus!");
		    Tag.this.cards.show(Tag.this, "p");
		    startEditing();
		}
	    });

	// Focus automatically starts/stops editing, but it has to be
	// signaled, too.
	tf.addFocusListener(new FocusAdapter() {
	    public void focusGained(FocusEvent e) {
		System.err.println("TF: focusGained");
		startEditing();
	    }
	    public void focusLost(FocusEvent e) {
		System.err.println("TF: focusLost");
		stopEditing();
	    }
	});

	tf.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		startEditing();
	    }

	    // A mouse press cancels an ongoing drag.
	    public void mousePressed(MouseEvent e) {
		System.err.println("Tag: mousePressed "+e.getX()+","+e.getY());
		if (dragInProgress) {
		    // Cancel drag
		    if (delDrag) {
			dragWindow.dispose();
			delDrag = false;
			System.err.println("Tag: DelDrag canceled");
		    }
		    dragInProgress = false;
		    System.err.println("Tag: Drag canceled");
		    return;
		}
	    }
      
	    // A drag ends, when the user releases the mouse. 
	    public void mouseReleased(MouseEvent e) {
		System.err.println("Tag mouseReleased "+e.getX()+","+e.getY());
		if (delDrag) {
		    // Did the deletion drag end inside the tag area
		    // or the tag that was dragged? If it did, the tag 
		    // won't be deleted.
		    Point p = new Point(e.getX(),e.getY());
		    Point q = new Point(e.getX(),e.getY());
		    SwingUtilities.convertPointToScreen(q, Tag.this);
		    SwingUtilities.convertPointFromScreen(q, tagArea);
		
		    if ( !Tag.this.contains(p) && !tagArea.contains(q)) {
			System.err.println("Tag: ...released outside...");
			tagArea.deleteTag(Tag.this);
		    }
		    dragWindow.dispose();
		    delDrag = false;
		}
		dragInProgress = false;
	    }

	    public void mouseEntered(MouseEvent e) {
		tagArea.setSelectedTag(Tag.this);
		stayVisible = true;
		setVisible(true);
		System.err.println("Tag: MouseEntered");
	    }

	    public void mouseExited(MouseEvent e) {
		tagArea.setSelectedTag(null);
		stayVisible = false;
		System.err.println("Tag: MouseExited");
	    }
	});
    
	// Drag'n'Drop stuff: the user may start the drag by dragging
	// the tag, not only the marker in the tag area.
	tf.addMouseMotionListener(new MouseMotionAdapter() {
	    public void mouseDragged(MouseEvent e) {
		System.err.println("Dragged "+e.getX()+","+e.getY());
		Point p = new Point(e.getX(),e.getY());

		// Drag starts only from inside the tag area. This
		// prevents the drag from restarting when it has been
		// canceled by chord click but the user still moves
		// the mouse == generates drag events.
		if ( !dragInProgress) {
		    if (Tag.this.contains(p)) {
			dragInProgress = true;
			System.err.println("Tag: Started Drag");
		    } else { // Not really dragging
			return;
		    }
		}
		
		// A deletion drag starts when the user has dragged
		// the tag outside a threshold area. If there were no
		// threshold, selecting text inside the label would be
		// impossible. 
		if ( !delDrag) {
		    Rectangle b = Tag.this.getBounds();
		    SwingUtilities.convertPointToScreen(p, Tag.this);
		    int xDist = Math.min(Math.abs(b.x - p.x), Math.abs(b.x + b.width - p.x));
		    int yDist = Math.min(Math.abs(b.y - p.y), Math.abs(b.y + b.height - p.y));

		    if (xDist < Tag.DELTHRESHOLD && yDist < Tag.DELTHRESHOLD) {
			// Don't start delete drag yet
			return;
		    }

		    delDrag = true;
		    dragWindow = cloneTagWindow();
		    // ((Tag) dragWindow).inDrag();
		} else {
		    SwingUtilities.convertPointToScreen(p, Tag.this);
		}

		p.x += 5;
		p.y += 5;
		dragWindow.setLocation(p);
	    }
	});

	// A DoucumentListener is used to prevent the JTextField from
	// scrolling. Now it grows instead.
	tf.getDocument().addDocumentListener(new DocumentListener() {
	    public void insertUpdate(DocumentEvent e) {
		updateTextfield();
	    }
	    public void removeUpdate(DocumentEvent e) {
		updateTextfield();
	    }
	    public void changedUpdate(DocumentEvent e) { }
	    public void updateTextfield() {
		Tag.this.pack();
	    }
	});

        // JSTT: Listens to enter presses, not used in this implementation
        // (except for a debug message) but somekind of functionality should
        // be implemented in the future (e.g. enter should unfocus the tag).
        tf.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
            // For compatibility with java.awt.TextField, the VK_ENTER key
            // fires the ActionEvent to the registered ActionListeners.
                System.err.println("Tag: Pressed enter");
            }
        });
    }                                               
    
    Tag(String newName, int newValue, TagArea newTagArea, Window topW) {
	this(newName, newValue, newTagArea, topW, true);
    }

    /**
     * Creates a copy of this tag to use as a drag image.
     */
    public Window cloneTagWindow() {
	Tag clone = new Tag(this.getText(),
			    this.getValue(),
			    this.tagArea,
			    this.topWindow, 
			    false);
	return clone;
    }

    /**
     * Sets the tag to the "normal state" (JTextField on top). This is 
     * used after the tag is cloned in order to use as a drag image.
     * The AWT textfield is by default on top first to gain the focus, 
     * which is never transfered to the drag image window.
     */
//     public void inDrag() {
// 	this.cards.show(this, "tf");
//     }

    public int getValue() { return this.value; }
    public void setValue(int newValue) {
	this.value = newValue;
    }

    int getDrawnY() { return this.drawnY; }
    void setDrawnY(int newDrawnY) {
	this.drawnY = newDrawnY;
	resetWindowLocation();
    }
  
    void resetWindowLocation() {
	Point upperLeftCorner = new Point(12,this.drawnY-10);
	// convert the point from component coordinates to screen coordinates
	// to determine absolute tooltip location.
	SwingUtilities.convertPointToScreen(upperLeftCorner,this.tagArea);
	this.setLocation(upperLeftCorner);
    }
  
    public String getText() { return this.tf.getText(); }
    public void setText(String newName) {
	this.tf.setText(newName);
	this.pack();
    }

    
    /**
     * Signals the listeners of a starting edit session and start the
     * edit.
     */
    public void startEditing() {
	if(editing)
	    return;
        editing = true;
	newBorn = false;
	System.err.println("Tag: startEditing()");
	fireTagEditStarted();

	this.cards.show(this,"tf");
	System.err.println("Tag: TF requests focus");
	this.tf.requestFocus();
    }
    
    /**
     * Signals the listeners of a the end of editing. 
     */
    protected void stopEditing() {
	if(!editing)
	    return;
	System.err.println("Tag: stopEditing()");
	fireTagEditStopped();
	editing = false;
    }

    // Listeners for Tags (Observer design pattern)
    //
    // The TagArea will add itself to listen all the tags in order to
    // catch their startEdit events. When startEdit happens, the
    // scroll should move to the corresponding tag. 
    
    EventListenerList listenerList = new EventListenerList();
    
    public void addTagListener(TagListener l) {
	listenerList.add(TagListener.class, l);
    }

    public void removeTagListener(TagListener l) {
	listenerList.remove(TagListener.class, l);
    }

    protected void fireTagEditStarted() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    TagEvent tagEvent = new TagEvent(this);
	    if (listeners[i]==TagListener.class) {
		((TagListener)listeners[i+1]).tagEditStarted(tagEvent);
	    }
	}
    }

    protected void fireTagEditStopped() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    TagEvent tagEvent = new TagEvent(this);
	    if (listeners[i]==TagListener.class) {
		((TagListener)listeners[i+1]).tagEditStopped(tagEvent);
	    }
	}
    }


    public void setVisible(boolean visibility) {
// 	System.err.println("Tag setVisible(" + visibility + "): stayVisible=" + stayVisible
// 			   + " edit=" + editing + " newBorn=" + newBorn);
	if (!visibility && (stayVisible || editing || newBorn))
	    return;

	if (visibility)
	    pack();
	else
	    setSize(new Dimension(5, 23));
    }

    public void setHighlight(boolean highlight) {
	if (highlight == true) {
	    this.tf.setBorder(hilightBorder);
	    this.tf.setBackground(hilightBg);
	} else {
	    this.tf.setBorder(basicBorder);
	    this.tf.setBackground(basicBg);
	}
    }
}
