/*
 * TagArea (c) GIST 1999
 */
  
package fi.helsinki.cs.gist.scroll;

import fi.helsinki.cs.gist.*;
import java.util.*;
import javax.swing.text.*;
import javax.swing.*;

import java.awt.*;
import java.awt.event.*;

import java.beans.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;

/**
 * TagArea is used in the Placeholder to display and keep track of
 * tags. 
 *
 * TagArea written by
 * @author asaura
 * @version 11.9.1999
 *
 */

class TagArea extends JComponent implements TagListener {

    /**
     * An ordered collection of tags. addTag() makes sure this
     * one stays ordered.
     */
    Vector tags = new Vector();

    /**
     * The tag area context menu. 
     */
    JPopupMenu popup;
  
    /**
     * The tag visibility state indicator. When tagsShown is true, the 
     * tags are visible.
     */
    boolean tagsShown;
    
    /**
     * The tag closest to the mouse pointer.
     */
    protected Tag tagCurrentlyUnderMouse = null;
    /**
     * The tag that will be operated with the context menu command.
     */
    protected Tag popupTag = null;
    /**
     * The tag that received the last mouse click. If a drag begins,
     * this is the tag that is going to be deleted.
     */
    protected Tag lastPressedTag = null;

    /**
     * NewTagHint icon (not selected).
     */
    Icon newTagHintIcon = null;

    /**
     * NewTagHint icon (selected).
     */
    Icon newTagHintIconButton = null;

    /**
     * NewTagHint Y position. 
     */
    int newTagHintY = -1;
    
    /**
     * NewTagHint document position. 
     */
    int newTagHintValue = -1;
    
    /**
     * True, when newTagHint is under the mouse.
     */
    boolean newTagHintSelected = false;
    
    /** How far from the center of the tag bullet can the mouse pointer
     *  be for the tag still to be selected (Y axis).
     */
    static int tagSelectionTreshold = 10;

    /**
     * Drag state information.
     */
    boolean dragInProgress = false;

    /**
     * The window that moves with the cursor during tag drag.
     */
    Window dragWindow = null;
  
    /**
     * The tag windows will placed in proportion to this window. When
       the top level window moves, the tag windows will be moved, too.
     */
    Window topLevelWindow;
  

    protected TagArea() {
	super();
	this.tagsShown = true;
	this.setLayout(null);
	this.popup = createPopup();
	
        newTagHintIcon = new ImageIcon(
		"fi/helsinki/cs/gist/images/EnabledNewTagIcon.gif");
        newTagHintIconButton = new ImageIcon(
                "fi/helsinki/cs/gist/images/NewTagIconButton.gif");

	this.addMouseListener(new MouseAdapter() {
	    public void mouseClicked(MouseEvent e) {
		System.err.println("TA: mouse clicked "+e.getY());
		Tag t = whichTagContains(e.getY());
		if(t == null)
		    return;
		if(e.getClickCount() == 2) { // doubleclick starts editing
		    TagArea.this.startEditing(t);
		}
// 		else { // single click requests thumb
// 		    tagEvent = new TagEvent(t);
// 		    fireTagActivated();
// 		}
	    }
      
	    public void mouseExited(MouseEvent e) {
		System.err.println("Exited "+e.getX()+","+e.getY());
		loseHighlight();
		tagCurrentlyUnderMouse = null;
                newTagHintSelected = false;
		repaint();
	    }
      
	    public void mouseEntered(MouseEvent e) {
		System.err.println("Entered "+e.getX()+","+e.getY());
	    }

	    // MousePressed may mean a drag cancel or a context menu.
	    public void mousePressed(MouseEvent e) {
		System.err.println("Pressed "+e.getX()+","+e.getY());
		if (dragInProgress) {
		    // Cancel drag
		    dragWindow.dispose();

		    // NB: dragInProgress should be set false here,
		    // but it's done now in maybeShowPopup, which is
		    // always called in mouseReleased. Otherwise we'd
		    // get a popup menu on mouse release. Oh bore... 
		    return;
		}

		lastPressedTag = whichTagContains(e.getY());
		maybeShowPopup(e);
		
                if (newTagHintSelected) {
		    addTag(newTagHintValue);
		} else { // request thumb
		    tagEvent = new TagEvent(lastPressedTag);
		    fireTagActivated();
		}

	    }
      
	    // The context menu is in here, too, because in some
	    // systems it pops up on mouse down, and in some systems
	    // on mouse up. MouseReleased ends also every drag, that
	    // has not been canceled.
	    public void mouseReleased(MouseEvent e) {
		System.err.println("Released "+e.getX()+","+e.getY());
		maybeShowPopup(e);
		Point p = new Point(e.getX(),e.getY());
		if(dragInProgress) {
		    if ( !contains(p)) {
			System.err.println("...released outside...");
			deleteTag(lastPressedTag);
		    }
		    dragWindow.dispose();
		}
		dragInProgress = false;
	    }
	});
    
	this.addMouseMotionListener(new MouseMotionAdapter() {
	    public void mouseMoved(MouseEvent e) {
                if (((e.getY()-tagSelectionTreshold) < newTagHintY) &&
		    (newTagHintY < (e.getY()+tagSelectionTreshold))) {
		    newTagHintSelected = true;
		} else {
		    newTagHintSelected = false;
		}
		
		Tag t = whichTagContains(e.getY());
		if (t != null && newTagHintSelected) {
		    // Cursor moved over newTagHint
		    tagCurrentlyUnderMouse = null;
		} else {
		    tagCurrentlyUnderMouse = t;
		}

		// Removed highligting to paint...

		// if (t != null)
		//     tagCurrentlyUnderMouse.setHighlight(true);

		repaint();
	    }

	    // Drag'n'Drop stuff. Starts the drag and sets the drag
	    // window to its correct location during drag.
	    public void mouseDragged(MouseEvent e) {
		System.err.println("Dragged "+e.getX()+","+e.getY());
		Point p = new Point(e.getX(),e.getY());

		if (lastPressedTag == null)
		    return; // No drag to start
		
		if(!contains(p))
		    System.err.println("...dragged outside...");
		
		// Start the drag, if it wasn't started yet.
		if(!dragInProgress) {
		    dragWindow = lastPressedTag.cloneTagWindow();
		    // ((Tag) dragWindow).inDrag();
		}

		// Set the drag window near the place where the drag
		// event was received.
		SwingUtilities.convertPointToScreen(p,TagArea.this);
		dragWindow.setLocation(p);
		dragInProgress = true;
	    }
	});

	// The correct value of the topLevelWindow is found out, when
	// a layout manager tells this component its size.
	this.addComponentListener(new ComponentAdapter() {
	    public void componentResized(ComponentEvent e) {
		System.out.println("TA: comp resized");
		if(topLevelWindow == null) {
		    Container cont = TagArea.this.getTopLevelAncestor();
		    if(cont instanceof Window) {
			topLevelWindow = (Window)cont;
		    }
		    else {
			System.err.println("Warning: Placeholder doesn't work right with applets");
		    }
		}
	    }
	});

	// When the top level window moves, the tag windows have to
	// move with it. 
	this.addAncestorListener(new AncestorListener() {
	    public void ancestorMoved(AncestorEvent event) {
		System.out.println("TA: Ancestor moved.");
		for(int i = 0; i < tags.size(); i++) {
		    Tag t = (Tag)tags.elementAt(i);
		    t.resetWindowLocation();
		}
	    }
	    public void ancestorAdded(AncestorEvent event) { }
	    public void ancestorRemoved(AncestorEvent event) { }
	});

    } // end constructor


    // Creates the context menu
    JPopupMenu createPopup() {
	JPopupMenu p = new JPopupMenu();
	JMenuItem menuItem;
	menuItem = new JMenuItem("Rename");
	menuItem.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		if (popupTag != null)
		    startEditing(popupTag);
	    }
	});
	p.add(menuItem);
	menuItem = new JMenuItem("Delete");
	menuItem.addActionListener(new ActionListener() {
	    public void actionPerformed(ActionEvent e) {
		if (popupTag != null) 
		    deleteTag(popupTag);
	    }
	});
	p.add(menuItem);
	return p;
    }

    /**
     * Moves to the nearest tag.
     */
    void snapToY(int y) {
	Tag t = whichTagContains(y);
	if(t != null) {
	    tagEvent = new TagEvent(t);
	    fireTagActivated();      
	}
    }
  
    /**
     * Displays the context menu, if this is the right time for it.
     */ 
    void maybeShowPopup(MouseEvent e) {
	if (e.isPopupTrigger()) {
	    // The drag is finally canceled here, see mousePressed for 
	    // explanations.
	    if (dragInProgress) {
		dragInProgress = false;
	    } else {
		popupTag = tagCurrentlyUnderMouse;
		if (popupTag != null)
		    popup.show(e.getComponent(), e.getX(), e.getY());
	    }
	}
    }

    /**
     * Displays the context menu for a given tag, if this is the right
     * time for it. 
     */ 
    void maybeShowPopup(MouseEvent e, Tag t) {
	if(e.isPopupTrigger()) {
	    popupTag = t;
	    popup.show(e.getComponent(), e.getX(), e.getY());
	}
    }
    
    void deleteTag(Tag t) {
	System.out.print("delete ");
	System.out.println(popupTag);
	tags.remove(t);
	t.dispose();
	repaint();
	tagEvent = new TagEvent(t);
	fireTagDeleted();
    }
  
    public void deleteAllTags() {
	ListIterator i = tags.listIterator();
	while (i.hasNext()) {
	    Tag t = (Tag) i.next();
	    t.dispose();
	    tagEvent = new TagEvent(t);
	    fireTagDeleted();
	}
	tags.removeAllElements();
    }

    public void setTags(Vector v) {
	deleteAllTags();
	ListIterator i = v.listIterator();
	while (i.hasNext()) {
	    TagModel m = (TagModel) i.next();
	    addTag(m.value, m.name, false); // Non-interactive add
	}
    }

    public Vector getTags() {
	Vector v = new Vector();
	ListIterator i = tags.listIterator();
	while (i.hasNext()) {
	    Tag t = (Tag) i.next();
	    TagModel m = new TagModel(t.getText(), t.getValue());
	    v.addElement(m);
	}
	return v;
    }

    void loseHighlight() {
	if(tagCurrentlyUnderMouse != null)
	    tagCurrentlyUnderMouse.setHighlight(false);
    }
  
    protected Tag whichTagContains(int y) {
	int distance = this.getSize().height;
	Tag t = null;
	for(int i = 0; i < tags.size(); i++) {
	    Tag tc = (Tag)tags.elementAt(i);
	    int dc = Math.abs(tc.getDrawnY() - y);
	    if(dc < distance) {
		distance = dc;
		t = tc;
	    }
	}
	return t; // return null if no tags
    }

    public boolean areaIntersectsTag(int start, int end) {
	for(int i = 0; i < tags.size(); i++) {
	    Tag t = (Tag)tags.elementAt(i);
	    int drawnLocation = t.getDrawnY();
	    if(drawnLocation > start && drawnLocation < end)
		return true;
	}
	return false;
    }
    

    protected void startEditing(Tag t) {
	// This is the only way move the focus from this window the
	// tag window. Swing sucks.
	t.p.requestFocus();
    }

    /**
     * Add an editable tag with the value given. Gain focus and start
     * editing.
     */
    protected Tag addTag(int newValue) {
	Tag newTag = addTag(newValue, "", true);
	startEditing(newTag);
	return newTag;
    }

    /**
      * Add a tag to its ordered place in the tags vector.
      */
    public Tag addTag(int newValue, String newName, boolean interactive) {
	final Tag newTag = new Tag(newName, newValue, this, this.topLevelWindow, interactive);
	tags.addElement(newTag);

	newTag.addTagListener(this);

	// TagArea shows the popup menus for tags
	newTag.tf.addMouseListener(new MouseAdapter() {
	    public void mousePressed(MouseEvent e) {
		Point p = new Point(e.getX(), e.getY());
		if (newTag.contains(p) && !newTag.dragInProgress)
		    maybeShowPopup(e, newTag);
	    }
	    public void mouseReleased(MouseEvent e) {
		Point p = new Point(e.getX(), e.getY());
		if (newTag.contains(p) && !newTag.dragInProgress)
		    maybeShowPopup(e, newTag);
	    }
	});

	// Set these variables, so that repaint paints the right thing
	// from the start
	if (newTagHintSelected) {
	    newTagHintSelected = false;
	    tagCurrentlyUnderMouse = newTag;
	}
	
	this.invalidate();
	this.revalidate();
	this.repaint();
	tagEvent = new TagEvent(newTag);
	fireTagAdded();
	return newTag;      
    }


    /**
     * Draws the TagArea and the Tags.
     */
    public void paintComponent(Graphics g) {
	g.setColor(new Color(220,220,220));
	Dimension d = this.getSize();
	g.fillRect(0,0,d.width,d.height);

	g.setColor(Color.black);
	for (int i = 0; i < tags.size(); i++) {
	    Tag t = (Tag)(tags.elementAt(i));
	    int drawnY = t.getDrawnY();
	    if (tagsShown || t == tagCurrentlyUnderMouse) {
 		t.setVisible(true);
		t.pack();
	    } else {
		t.setVisible(false);
	    }
	    g.drawLine(5,drawnY,10,drawnY);
	    g.fillOval(4,drawnY-2,5,5);
	    
	    if(t == tagCurrentlyUnderMouse) {
		g.drawOval(2,drawnY-4,9,9);
		t.setHighlight(true);
	    }
	    else {
		t.setHighlight(false);
	    }
	}

	// Paint newTagHint, if needed
        if ((newTagHintY != -1) && (newTagHintIcon != null))
            if (newTagHintSelected)
                newTagHintIconButton.paintIcon(this, g, 1, newTagHintY - ((newTagHintIconButton.getIconHeight())/2) );
            else
                newTagHintIcon.paintIcon(this, g, 3, newTagHintY - ((newTagHintIcon.getIconHeight())/2) );
    }

    /**
     * Sets the place of the newTagIcon and causes a repaint.
     */
    public void setNewTagHint(int drawY, int value) {
        newTagHintY = drawY;
        newTagHintValue = value;
        repaint();
    }

    /**
      * Show/hide tags.
      */
    public void setTagsVisible(boolean visibility) {
	tagsShown = visibility;
	repaint();
	if (tagsShown)
	    fireTagsShown();
	else
	    fireTagsHidden();
    }

    /**
      * Sets a tags as selected tag.
      */
    public void setSelectedTag(Tag t) {
	tagCurrentlyUnderMouse = t;
	repaint();
    }
    
    // Listeners for TagArea
    //
    // The TagArea generates TagEvents, whenever a tag is added,
    // deleted or clicked. At least the scrollbar should listen to
    // the thumb request and move the thumb to the corresponding tag. 
    
    EventListenerList listenerList = new EventListenerList();
    TagEvent tagEvent = null;

    public void addTagAreaListener(TagAreaListener l) {
	listenerList.add(TagAreaListener.class, l);
    }

    public void removeTagAreaListener(TagAreaListener l) {
	listenerList.remove(TagAreaListener.class, l);
    }

    protected void fireTagAdded() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagAdded(tagEvent);
	    }
	}
    }
    protected void fireTagDeleted() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagDeleted(tagEvent);
	    }
	}
    }
    protected void fireTagChanged() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagChanged(tagEvent);
	    }
	}
    }
    protected void fireTagActivated() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagActivated(tagEvent);
	    }
	}
    }
    protected void fireTagsShown() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagsShown(tagEvent);
	    }
	}
    }
    protected void fireTagsHidden() {
	Object[] listeners = listenerList.getListenerList();
	for (int i = listeners.length-2; i>=0; i-=2) {
	    if (listeners[i]==TagAreaListener.class) {
		if (tagEvent == null)
		    return;
		((TagAreaListener)listeners[i+1]).tagsHidden(tagEvent);
	    }
	}
    }


    // TagListener implementation
    //
    // The Tags fire events whenever their editing starts of stops.
    // The TagArea is a listener for all of its tags.
    
    public void tagEditStarted(TagEvent e) {
	System.err.println("PUI: tag edit start event received");
	tagEvent = e;
	fireTagActivated();
    }

    public void tagEditStopped(TagEvent e) {
	System.err.println("PUI: tag edit stop event received");
	requestFocus();
    }

}

            
            
